def edit_distance(s: str, t: str, maxdiff: int = ...) -> int: ...
def edit_distance_affine_gap(
    query: str, ref: str, mismatch_cost: int, gap_start: int = ..., gap_extend: int = ...
) -> int: ...
